import logging

import click
from scrapy.utils.python import garbage_collect

logger = logging.getLogger(__name__)


@click.group()
def main():
    click.echo("cli @todo")


@click.command()
def crawl():
    print("crawl")


main.add_command(crawl)

if __name__ == "__main__":
    try:
        main()
    except KeyboardInterrupt:
        logger.error("User interrupted")
    except Exception as e:
        logger.error(e)
    finally:
        garbage_collect()




from opennem.spiders.nemweb import NemwebSpider


class NemwebCurrentDispatchIS(NemwebSpider):
    name = "au.nem.current.dispatch_is"
    start_urls = ["http://nemweb.com.au/Reports/Current/DispatchIS_Reports/"]
    limit = 0


class NemwebArchiveDispatchIS(NemwebSpider):
    name = "au.nem.archive.dispatch_is"
    start_urls = ["http://nemweb.com.au/Reports/Archive/DispatchIS_Reports/"]
    limit = 0

    # Archives tend to contain large zips of embedded zips so throttle
    # to limit memory use
    custom_settings = {
        'CONCURRENT_REQUESTS': 1,
        'CONCURRENT_ITEMS': 1,
    }



"""scada_dispatch

Revision ID: 01d74b813948
Revises: 
Create Date: 2020-06-09 17:47:32.424246

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "01d74b813948"
down_revision = None
branch_labels = None
depends_on = None


def upgrade(engine_name):
    globals()["upgrade_%s" % engine_name]()


def downgrade(engine_name):
    globals()["downgrade_%s" % engine_name]()


def upgrade_opennem():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "nem_dispatch_unit_scada",
        sa.Column(
            "created_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column("updated_at", sa.DateTime(timezone=True), nullable=True),
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("SETTLEMENTDATE", sa.DateTime(), nullable=True),
        sa.Column("DUID", sa.Text(), nullable=True),
        sa.Column(
            "SCADAVALUE", sa.NUMERIC(precision=10, scale=6), nullable=True
        ),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_index(
        op.f("ix_nem_dispatch_unit_scada_DUID"),
        "nem_dispatch_unit_scada",
        ["DUID"],
        unique=False,
    )
    op.create_index(
        op.f("ix_nem_dispatch_unit_scada_SETTLEMENTDATE"),
        "nem_dispatch_unit_scada",
        ["SETTLEMENTDATE"],
        unique=False,
    )
    op.create_index(
        "nem_dispatch_unit_scada_uniq",
        "nem_dispatch_unit_scada",
        ["SETTLEMENTDATE", "DUID"],
        unique=True,
    )
    # ### end Alembic commands ###


def downgrade_opennem():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index(
        "nem_dispatch_unit_scada_uniq", table_name="nem_dispatch_unit_scada"
    )
    op.drop_index(
        op.f("ix_nem_dispatch_unit_scada_SETTLEMENTDATE"),
        table_name="nem_dispatch_unit_scada",
    )
    op.drop_index(
        op.f("ix_nem_dispatch_unit_scada_DUID"),
        table_name="nem_dispatch_unit_scada",
    )
    op.drop_table("nem_dispatch_unit_scada")
    # ### end Alembic commands ###




import logging
import re

import scrapy
from scrapy import Spider

from opennem.datetimes import parse_date

PADDING_WIDTH = 7

__is_number = re.compile(r"^\d+$")


def is_number(value):
    if re.match(__is_number, value):
        return True
    return False


def parse_dirlisting(raw_string):
    """
        given a raw text directory listing like "     Saturday 11th June 2020      6789"
        will parse and return both the date and listing type

        @param raw_string - the raw directory listing string
        @return dict of the date in iso format and the type (file or directory)
    """
    components = raw_string.split(" " * PADDING_WIDTH)
    components = [i.strip() for i in components]
    components = list(filter(lambda x: x != "", components))

    _ltype = "dir"

    if is_number(components[1]):
        _ltype = "file"

    return {
        "date": parse_date(components[0]).isoformat(),
        "type": _ltype,
    }

class DirlistingSpider(Spider):
    """
        spider that parses html directory listings produced by web servers


    """
    limit = 0

    def parse(self, response):
        links = [i.get() for i in response.xpath("//body/pre/br/following-sibling::a/@href")]
        metadata = [parse_dirlisting(i.get()) for i in response.xpath("//body/pre/br/following-sibling::text()")]

        parsed = 0

        for i, entry in enumerate(metadata):
            if entry["type"] == "file":
                link = response.urljoin(links[i])

                self.log("Getting {}".format(link), logging.INFO)

                if self.limit and self.limit > 0 and (parsed >= self.limit):
                    self.log(f"Reached limit of {self.limit}", logging.INFO)
                    return None

                parsed += 1

                yield from self.parse_entry({
                    "link": link,
                    **entry
                })

    def parse_entry(self, entry):
        yield entry



"""
    OpenNEM primary schema adapted to support multiple energy sources

    Currently supported:

    - NEM
    - WEM
"""

from sqlalchemy import (
    NUMERIC,
    Column,
    Date,
    DateTime,
    ForeignKey,
    Index,
    Integer,
    Sequence,
    String,
    Table,
    Text,
    Time,
)
from sqlalchemy.dialects import postgresql
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import relationship
from sqlalchemy.sql import func

Base = declarative_base()
metadata = Base.metadata


class NemModel(object):
    """
        Base model for both NEM and WEM

        Each table has an overrid for update and additional meta fields

        @TODO - upsert support for postgresql and mysql dialects (see db/__init__)
    """

    def update(self, **kwargs):
        for key, value in kwargs.items():
            if key == "id" or key.endswith("_id"):
                continue

            # @TODO watch how we do updates so we don't overwrite data
            cur_val = getattr(self, key)
            if key not in [None, ""]:
                setattr(self, key, value)

    def __compile_query(self, query):
        """Via http://nicolascadou.com/blog/2014/01/printing-actual-sqlalchemy-queries"""
        compiler = (
            query.compile
            if not hasattr(query, "statement")
            else query.statement.compile
        )
        return compiler(dialect=postgresql.dialect())

    def __upsert(
        self,
        session,
        model,
        rows,
        as_of_date_col="report_date",
        no_update_cols=[],
    ):
        table = model.__table__

        stmt = self.insert(table).values(rows)

        update_cols = [
            c.name
            for c in table.c
            if c not in list(table.primary_key.columns)
            and c.name not in no_update_cols
        ]

        on_conflict_stmt = stmt.on_conflict_do_update(
            index_elements=table.primary_key.columns,
            set_={k: getattr(stmt.excluded, k) for k in update_cols},
            index_where=(model.report_date < stmt.excluded.report_date),
        )

        print(self.compile_query(on_conflict_stmt))
        session.execute(on_conflict_stmt)

    created_at = Column(DateTime(timezone=True), server_default=func.now())
    updated_at = Column(DateTime(timezone=True), onupdate=func.now())


class NemDispatchUnitScada(Base, NemModel):
    __tablename__ = "nem_dispatch_unit_scada"
    __table_args__ = (
        Index(
            "nem_dispatch_unit_scada_uniq",
            "SETTLEMENTDATE",
            "DUID",
            unique=True,
        ),
    )

    id = Column(Integer, primary_key=True)
    SETTLEMENTDATE = Column(DateTime, index=True)
    DUID = Column(Text, index=True,)
    SCADAVALUE = Column(NUMERIC(10, 6))


class NemDispatchPrice(Base, NemModel):
    __tablename__ = "nem_dispatch_price"
    __table_args__ = (
        Index(
            "nem_dispatch_price_uniq", "SETTLEMENTDATE", "DUID", unique=True,
        ),
    )

    id = Column(Integer, primary_key=True)
    SETTLEMENTDATE = Column(DateTime, index=True)
    DUID = Column(Text, index=True,)
    SCADAVALUE = Column(NUMERIC(10, 6))

    REGIONID = Column(Text, index=True,)
    RUNO = Column(Integer, nullable=False)

    DISPATCHINTERVAL = Column(Integer, nullable=False)

    INTERVENTION = Column(Integer, nullable=False)

    RRP = Column(Integer, nullable=False)

    EEP = Column(Integer, nullable=False)

    ROP = Column(Integer, nullable=False)

    APCFLAG = Column(Integer, nullable=False)

    MARKETSUSPENDEDFLAG = Column(Integer, nullable=False)

    LASTCHANGED = Column(Integer, nullable=False)

    RAISE6SECRRP = Column(Integer, nullable=False)

    RAISE6SECROP = Column(Integer, nullable=False)

    RAISE6SECAPCFLAG = Column(Integer, nullable=False)

    RAISE60SECRRP = Column(Integer, nullable=False)

    RAISE60SECROP = Column(Integer, nullable=False)

    RAISE60SECAPCFLAG = Column(Integer, nullable=False)

    RAISE5MINRRP = Column(Integer, nullable=False)

    RAISE5MINROP = Column(Integer, nullable=False)

    RAISE5MINAPCFLAG = Column(Integer, nullable=False)

    RAISEREGRRP = Column(Integer, nullable=False)

    RAISEREGROP = Column(Integer, nullable=False)

    RAISEREGAPCFLAG = Column(Integer, nullable=False)

    LOWER6SECRRP = Column(Integer, nullable=False)

    LOWER6SECROP = Column(Integer, nullable=False)

    LOWER6SECAPCFLAG = Column(Integer, nullable=False)

    LOWER60SECRRP = Column(Integer, nullable=False)

    LOWER60SECROP = Column(Integer, nullable=False)

    LOWER60SECAPCFLAG = Column(Integer, nullable=False)

    LOWER5MINRRP = Column(Integer, nullable=False)

    LOWER5MINROP = Column(Integer, nullable=False)

    LOWER5MINAPCFLAG = Column(Integer, nullable=False)

    LOWERREGRRP = Column(Integer, nullable=False)

    LOWERREGROP = Column(Integer, nullable=False)

    LOWERREGAPCFLAG = Column(Integer, nullable=False)

    PRICE_STATUS = Column(Integer, nullable=False)

    PRE_AP_ENERGY_PRICE = Column(Integer, nullable=False)

    PRE_AP_RAISE6_PRICE = Column(Integer, nullable=False)

    PRE_AP_RAISE60_PRICE = Column(Integer, nullable=False)

    PRE_AP_RAISE5MIN_PRICE = Column(Integer, nullable=False)

    PRE_AP_RAISEREG_PRICE = Column(Integer, nullable=False)

    PRE_AP_LOWER6_PRICE = Column(Integer, nullable=False)

    PRE_AP_LOWER60_PRICE = Column(Integer, nullable=False)

    PRE_AP_LOWER5MIN_PRICE = Column(Integer, nullable=False)

    PRE_AP_LOWERREG_PRICE = Column(Integer, nullable=False)

    CUMUL_PRE_AP_ENERGY_PRICE = Column(Integer, nullable=False)

    CUMUL_PRE_AP_RAISE6_PRICE = Column(Integer, nullable=False)

    CUMUL_PRE_AP_RAISE60_PRICE = Column(Integer, nullable=False)

    CUMUL_PRE_AP_RAISE5MIN_PRICE = Column(Integer, nullable=False)

    CUMUL_PRE_AP_RAISEREG_PRICE = Column(Integer, nullable=False)

    CUMUL_PRE_AP_LOWER6_PRICE = Column(Integer, nullable=False)

    CUMUL_PRE_AP_LOWER60_PRICE = Column(Integer, nullable=False)

    CUMUL_PRE_AP_LOWER5MIN_PRICE = Column(Integer, nullable=False)

    CUMUL_PRE_AP_LOWERREG_PRICE = Column(Integer, nullable=False)

    OCD_STATUS = Column(Integer, nullable=False)

    MII_STATUS = Column(Integer, nullable=False)


class WemFacilityScada(Base, NemModel):
    __tablename__ = "wem_facility_scada"
    __table_args__ = (
        Index(
            "wem_facility_scada_uniq",
            "TRADING_INTERVAL",
            "FACILITY_CODE",
            unique=True,
        ),
    )

    id = Column(Integer, primary_key=True)
    TRADING_INTERVAL = Column(DateTime, index=True)
    PARTICIPANT_CODE = Column(Text, index=True,)
    FACILITY_CODE = Column(Text, index=True,)
    ENERGY_GENERATED = Column(NUMERIC(10, 6))
    EOI_QUANTITY = Column(NUMERIC(10, 6))




import logging

import scrapy

from opennem.pipelines.files import LinkExtract
from opennem.pipelines.nem import DatabaseStore, ExtractCSV
from opennem.spiders.dirlisting import DirlistingSpider


class WemSpider(DirlistingSpider):
    allowed_domains = ["wa.nemweb.com.au"]
    pipelines = set([
        LinkExtract,
        ExtractCSV,
        DatabaseStore,
    ])



"""
Scrapy settings for nemweb project

For simplicity, this file contains only settings considered important or
commonly used. You can find more settings consulting the documentation:

https://docs.scrapy.org/en/latest/topics/settings.html
"""

BOT_NAME = "opennem"
SPIDER_MODULES = ["opennem.spiders"]
NEWSPIDER_MODULE = "opennem.spiders"


USER_AGENT = "opennem (+https://opennem.org.au)"
ROBOTSTXT_OBEY = True
CONCURRENT_REQUESTS = 16
COOKIES_ENABLED = False
TELNETCONSOLE_ENABLED = False

DEFAULT_REQUEST_HEADERS = {
    "Accept": "text/html,application/xhtml+xml,application/xml;q=0.9,*/*;q=0.8",
    "Accept-Language": "en",
}

AUTOTHROTTLE_ENABLED = True
AUTOTHROTTLE_START_DELAY = 5
AUTOTHROTTLE_MAX_DELAY = 60


# Pipline docs:
# 1xx series - download handlers
# 2xx series - unpackers
# 3xx series -
# 4xx series - database processes
ITEM_PIPELINES = {
    "opennem.pipelines.files.LinkExtract": 101,
    "opennem.pipelines.nem.ExtractCSV": 301,
    "opennem.pipelines.nem.DatabaseStore": 401,
}

# @TODO if DEBUG
HTTPCACHE_ENABLED = False
HTTPCACHE_DIR = ".scrapy"



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['opennem',
 'opennem.agent',
 'opennem.commands',
 'opennem.datetimes',
 'opennem.db',
 'opennem.db.migrations',
 'opennem.db.migrations.versions',
 'opennem.db.models',
 'opennem.nem_derived',
 'opennem.pipelines',
 'opennem.schema',
 'opennem.settings',
 'opennem.spiders',
 'opennem.spiders.nem',
 'opennem.spiders.wem',
 'opennem.utils']

package_data = \
{'': ['*']}

install_requires = \
['alembic>=1.4.2,<2.0.0',
 'click>=7.1.2,<8.0.0',
 'pydantic>=1.5.1,<2.0.0',
 'python-dotenv>=0.13.0,<0.14.0',
 'pyyaml>=5.3.1,<6.0.0',
 'requests>=2.23.0,<3.0.0',
 'scrapy>=2.1.0,<3.0.0',
 'sentry-sdk>=0.14.4,<0.15.0',
 'smart_open>=2.0.0,<3.0.0',
 'sqlalchemy>=1.3.17,<2.0.0']

entry_points = \
{'console_scripts': ['bluestoned = opennem.cli:main']}

setup_kwargs = {
    'name': 'opennem',
    'version': '0.3.0',
    'description': 'opennem engine agent',
    'long_description': '# OpenNEM Energy Market Data Access\n\nThe OpenNEM project aims to make the wealth of public National Electricity Market (NEM) data more accessible to a wider audience.\n\nThis toolkit enables downloading, mirroring and accessing energy data from various networks\n\nProject homepage at https://opennem.org.au\n\nAvailable on Docker at https://hub.docker.com/r/opennem/opennem\n\nCurrently supports:\n\n- Australian NEM: https://www.nemweb.com.au/\n- West Australia Energy Market: http://data.wa.aemo.com.au/\n\n\n## Install\n\nYou can install this project with python `pip`:\n\n```sh\n$ pip install opennem\n```\n\nOr alternatively with docker:\n\n```\n$ docker pull opennem/opennem\n```\n\nBundled with sqlite support. Other database drivers are optional and not installed by default. Install a supported database driver:\n\nPostgres:\n\n```sh\n$ pip install psycopg2\n```\n\n## Usage\n\nList the crawlers\n\n```sh\n$ scrapy list\n```\n\nCrawl\n\n```sh\n$ scrapy crawl au.nem.current.dispatch_scada\n```\n\n## Development\n\nSetting up a virtual environment and installing requiements using Poetry:\n\n```sh\n$ poetry install\n$ source .venv/bin/activate\n```\n\nSettings are read from environment variables. Environment variables can be read from a `.env` file in the root of the folder. Setup the environment by copying the `.env.sample` file to `.env`. The defaults in the sample file map to the settings in `docker-compose.yml`\n\nThere is a `docker-compose` file that will bring a local database:\n\n```sh\n$ docker-compose up -d\n```\n\nBring up the database migrations using alembic:\n\n```sh\n$ alembic upgrade head\n```\n\nRun scrapy in the root folder for options:\n\n```sh\n$ scrapy\n```\n\nThe `opennem` cli interface provides other options and settings:\n\n```sh\n$ opennem -h\n```\n\nSettings for Visual Studio Code are stored in `.vscode`. Code is kept formatted and linted using `pylint`, `black` and `isort` with settings defined in `pyproject.toml`\n\n## Architecture overview\n\nThis project uses [Scrapy](https://scrapy.org/) to obtain data from supported energy markets and [SQLAlchemy](https://www.sqlalchemy.org/) to store data, and [Alembic](https://alembic.sqlalchemy.org/en/latest/) for database migrations. Database storage has been tested with sqlite, postgres and mysql.\n\nOverview of scrapy architecture:\n\n![](https://docs.scrapy.org/en/latest/_images/scrapy_architecture_02.png)\n\n## Code Navigation\n\n* Spider definitions in `opennem/spiders`\n* Processing pipelines for crawls in `opennem/pipelines`\n* Database models for supported energy markets are stored in `opennem/db/models`\n',
    'author': 'Dylan McConnell',
    'author_email': 'dylan.mcconnell@unimelb.edu.au',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://opennem.org.au',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)




from opennem.spiders.wem_base import WemSpider


class NemwebCurrentDispatchScada(WemSpider):
    name = "au.wem.current.facility_scada"
    start_urls = ["http://data.wa.aemo.com.au/datafiles/facility-scada/"]
    limit = 0



import logging
import os

from sqlalchemy import create_engine
from sqlalchemy.ext.compiler import compiles
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import sessionmaker
from sqlalchemy.sql.expression import Insert

from opennem.settings import get_database_host, get_mysql_host

DeclarativeBase = declarative_base()

logger = logging.getLogger(__name__)


# @compiles(Insert, "postgresql")
def compile_upsert(insert_stmt, compiler, **kwargs):
    """
    converts every SQL insert to an upsert  i.e;
    INSERT INTO test (foo, bar) VALUES (1, 'a')
    becomes:
    INSERT INTO test (foo, bar) VALUES (1, 'a') ON CONFLICT(foo) DO UPDATE SET (bar = EXCLUDED.bar)
    (assuming foo is a primary key)
    :param insert_stmt: Original insert statement
    :param compiler: SQL Compiler
    :param kwargs: optional arguments
    :return: upsert statement
    """
    pk = insert_stmt.table.primary_key
    insert = compiler.visit_insert(insert_stmt, **kwargs)
    ondup = f'ON CONFLICT ({",".join(c.name for c in pk)}) DO UPDATE SET'
    updates = ", ".join(
        f"{c.name}=EXCLUDED.{c.name}" for c in insert_stmt.table.columns
    )
    upsert = " ".join((insert, ondup, updates))
    return upsert


def db_connect(db_name=None, debug=False):
    """
    Performs database connection using database settings from settings.py.
    Returns sqlalchemy engine instance
    """
    db_conn_str = get_database_host()

    logger.info("Connecting with {}".format(db_conn_str))

    try:
        e = create_engine(db_conn_str, echo=debug)
        return e
    except Exception as e:
        logger.error("Could not connect to database: {}".format(e))
        return



import functools
import logging
import re

import six


def regex(x):
    if isinstance(x, six.string_types):
        return re.compile(x)
    return x


def check_spider_pipeline(process_item_method):
    @functools.wraps(process_item_method)
    def wrapper(self, item, spider):

        # message template for debugging
        msg = "%%s %s pipeline step" % (self.__class__.__name__,)

        if not hasattr(spider, "pipelines"):
            spider.pipelines = set()

        if self.__class__ in spider.pipelines:
            spider.log(msg % "executing", level=logging.DEBUG)
            return process_item_method(self, item, spider)

        else:
            spider.log(msg % "skipping", level=logging.DEBUG)
            return item

    return wrapper
